/**
 * The operands of all bitwise operators are converted to signed 32-bit integers
 * in big-endian order and in two's complement format. Big-endian order means
 * that the most significant bit (the bit position with the greatest value) is
 * the left-most bit if the 32 bits are arranged in a horizontal line. Two's
 * complement format means that a number's negative counterpart (e.g. 5 vs. -5)
 * is all the number's bits inverted (bitwise NOT of the number, a.k.a. one's
 * complement of the number) plus one.
 */

define([
  '../../core/Class',
  '../../mixins/NumberCapsule'
],
function (Class, NumberCapsule) {
  'use strict';

  return Class.extend('Address', [new NumberCapsule('address')], {

    /**
     * Create an IP address
     *
     * @param {String/Number} address
     * @constructor
     */
    init: function (address) {

      // Check parameter
      if(typeof address !== 'string' && typeof address !== 'number') {
        throw new Error('Expecting parameter to be a String/Number, "' + (typeof address) + '" given.');
      }

      // If the address is a number, consider it as the IP Address number representation
      if(/^-?[0-9]+$/.test(address)) {

        this.setAddress(Number(address));

        if(this.getAddress() < -2147483648 || this.getAddress() > 2147483647) {
          throw new Error('Invalid IPv4 address "' + this.getAddress() + '"');
        }

      }
      // Otherwise, parse it as a string
      else {

        // Retrieve the different part
        var ip = address.match && address.match(/^(\d+)\.(\d+)\.(\d+)\.(\d+)$/);

        if(!ip || ip[1] > 255 || ip[2] > 255 || ip[3] > 255 || ip[4] > 255) {
          throw new Error('Invalid IPv4 Address "' + address + '"');
        }

        // Calculate the IP address number
        this.setAddress( (+ip[1]<<24) + (+ip[2]<<16) + (+ip[3]<<8) + (+ip[4]) );
      }
    },

    /**
     * Convert an {@link Address IPv4 Address} into a String in the following
     * format:
     *
     *     "ABC.DEF.GHI.JKL"
     *
     * @return {String} Returns the standard String representation of the {@link Address IPv4 Address}
     */
    toString: function () {
      return  ((this.getAddress() >>> 24) % 256) +
        '.' + ((this.getAddress() >>> 16) % 256) +
        '.' + ((this.getAddress() >>> 8 ) % 256) +
        '.' + ((this.getAddress() >>> 0 ) % 256);
    },

    /**
     * Get the next IP address
     *
     * @return {Address} Returns the next IP address
     */
    getNext: function() {

      if(this.getAddress() === 2147483647) { // 0111..111 => 1000..000
        return new this.constructor(-2147483648);
      }

      if(this.getAddress() === -1) { // 255.255.255.255
        throw new Error('IP address 255.255.255.255 has no next address');
      }

      return new this.constructor(this.getAddress() + 1);
    },

    /**
     * Get the previous IP address
     *
     * @return {Address} Returns the previous IP address
     */
    getPrevious: function() {

      if(this.getAddress() === -2147483648) { // 1000..000 => 0111..111
        return new this.constructor(2147483647);
      }

      if(this.getAddress() === 0) { // 0.0.0.0
        throw new Error('IP address 0.0.0.0 has no previous address');
      }

      return new this.constructor(this.getAddress() - 1);
    }

  });
});
