define([
  '../../core/Class',
  '../../mixins/NumberCapsule',
  './Address'
],
function (Class, NumberCapsule, Address) {
  'use strict';

  return Class.extend('Mask', [new NumberCapsule('mask')], {

    /**
     * Create an IP mask
     *
     * @param {Number/String/Address} mask The mask size (Ex: 24, '24') or
     * address (Ex: '255.255.255.0', Address('255.255.255.0'))
     * @constructor
     */
    init: function(mask) {

      // Mask size (Ex: 24)
      if(/^[0-9]+$/.test(mask)) {

        // Retrieve the mask size (31 to 0)
        var maskSize = Number(mask);
        if(maskSize > 32) {
          throw new Error('Expecting mask size between 0 and 32, got ' + maskSize);
        }

        // Calculate the actual mask (-xxxx)
        // -1 is the full "ones" number, aka 1111 1111 1111 1111 in 16-bits
        this.setMask(maskSize === 0 ? 0 : (-1 << (32 - maskSize)));
      }
      // Address (Ex: 255.255.255.0)
      else {
        this.setMask((mask instanceof Address ? mask : new Address(mask)).address);
      }
    },

    /**
     * Compute the mask size
     *
     * @return {Number} Returns the mask size
     */
    getMaskSize: function() {

      /**
       * mask        = 1111 1111 1111 0000
       * ~(mask)     = 0000 0000 0000 1111
       * ~(mask) + 1 = 0000 0000 0001 0000 = Math.pow(2, 32 - maskSize)
       *
       * a = Math.pow(b,c)  <=>  c = Math.log(a)/Math.log(b)
       *
       * a = Math.pow(2, 32 - maskSize)
       * b = 2
       * c = 32 - maskSize
       *
       * ~(mask) + 1 = Math.pow(2, 32 - maskSize)  <=> 32 - maskSize = Math.log(~(this.mask) + 1)/Math.log(2)
       *                                           <=> maskSize = 32 - Math.log(~(this.mask) + 1)/Math.log(2)
       */
      switch(this.getMask()) {
        case -1: return 32; // Prevents returning Infinity (see Math.log)
        case  0: return 0; // Prevents returning 32
        default: return Math.round(32 - Math.log(~(this.getMask()) + 1)/Math.log(2));
      }
    },

    apply: function(address) {
      return new Address(this.getMask() & (address instanceof Address ? address : new Address(address)).address);
    },

    toString: function() {
      return new Address(this.getMask()).toString();
    }

  });
});
