define([
  '../../core/Class',
  '../../mixins/NumberCapsule',
  '../../mixins/ArrayCapsule',
  './Address',
  './Mask',
  './Subnet'
],
function (Class, TypedCapsule, ArrayCapsule, Address, Mask, Subnet) {
  'use strict';

  return Class.extend('Pool', [
    new TypedCapsule('subnet', Subnet),
    new ArrayCapsule('allocatedAddresses')
  ], {

    /**
     * IP v4 address Pool
     *
     * An IP address pool that can be defined from a subnet and can manage allocated
     * IP addresses.
     *
     *     new Subnet('192.168.1.0', 24).getPool();
     *
     * @param subnetOrNetwork
     * @class Pool
     */
    init: function(subnet) {

      // Subnet check
      if(!subnet || !(subnet instanceof Subnet)) {
        throw new Error('Expecting parameter to be a Subnet, ' + subnet + ' given.');
      }

      /**
       * @type {Subnet} The pool's subnet
       */
      this.setSubnet(subnet);

      /**
       * @type {Address[]} Allocated addresses
       */
      this.setAllocatedAddresses([]);

    },

    /**
     * Determine whether an IP address is allocated or not. An allocated IP address
     * is one of the IP address allocated within the pool using
     * {@link #allocate the allocation method}.
     *
     * @param {String/Number/Address} address The address to test
     * @return {Boolean} Return true if the address is allocated, false otherwise
     */
    isAllocated: function(address) {

      address = address instanceof Address ? address : new Address(address);

      if(!this.getSubnet().isValidAddress(address)) {
        throw new Error('The IP address ' + address + ' does not belong to the subnet ' + this.getSubnet());
      }

      for(var i = 0 ; i < this.allocatedAddresses.length ; i++) {
        if(this.getAllocatedAddresses()[i].getAddress() === address.getAddress()) { // TODO equals
          return true;
        }
      }

      return false;
    },

    /**
     * Determine whether an IP address is available or not. An available IP address
     * is must match the two following conditions:
     *
     *    - the address is {@link Subnet#isValidAddress valid within the subnet}
     *    - the address is not {@link #isAllocated allocated}
     *
     * @param {String/Number/Address} address The address to test
     * @return {Boolean} Return true if the address is allocated, false otherwise
     */
    isAvailable: function(address) {

      address = address instanceof Address ? address : new Address(address);

      return this.getSubnet().isValidAddress(address) && !this.isAllocated(address);
    },

    /**
     * Allocate an IP address
     *
     * @param {String/Number/Address/String[]/Number[]/Address[]} address The address(es) to allocate
     */
    allocate: function(address) {
      var toString = String.prototype.toString;

      // If an array, allocate all the IP addresses
      if(Object.prototype.toString.call(address) === "[object Array]") {
        for(var i = 0 ; i < address.length ; i++) {

          var value = address[i];

          // Prevent using nested arrays
          if(Object.prototype.toString.call(value) === "[object Array]") {
            throw new Error('Nested arrays are not supported by Pool#allocate()');
          }

          this.allocate(value);
        }
      }
      // Otherwise, allocate the address
      else {
        address = address instanceof Address ? address : new Address(address);

        if(!this.isAllocated(address)) {
          this.allocatedAddresses.push(address);
        }
      }

      return this;
    },

    /**
     * Get the first available address
     *
     * @return {Address} Return the first available address, or `null` if
     * none is available
     */
    getFirstAvailable: function() {

      var subnet = this.getSubnet(),
          first;

      for(first = subnet.getNetwork().getNext() ; first !== null && subnet.isValidAddress(first) && this.isAllocated(first) ; first = first.getNext()) {
        // Do nothing
      }

      return subnet.isValidAddress(first) ? first : null;
    },

    /**
     * Get the last available address
     *
     * @return {Address} Return the last available address, or `null` if
     * none is available
     */
    getLastAvailable: function() {

      var subnet = this.getSubnet(),
          last;

      for(last = subnet.getBroadcastAddress().getPrevious() ; last !== null && subnet.isValidAddress(last) && this.isAllocated(last) ; last = last.getPrevious()) {
        // Do nothing
      }

      return subnet.isValidAddress(last) ? last : null;
    }


  });
});
