define([
  '../../core/Class',
  '../../mixins/TypedCapsule',
  './Address',
  './Mask'
],
function (Class, TypedCapsule, Address, Mask) {
  'use strict';

  var Subnet = Class.extend('Subnet', [
      new TypedCapsule('network', Address),
      new TypedCapsule('mask'   , Mask)
    ], {

      /**
       * Get an {@link Address address} as an object
       *
       * @param {String/Number/Address} address The source address
       *
       * @return {Address} Returns a new {@link Address} object if `address`
       * is a `String` or a `Number`, or the original {@link Address} object otherwise
       */
      getAddressAsObject: function(address) {
        return address instanceof Address ? address : new Address(address);
      },

      /**
       * Get a {@link Mask mask} as an object
       *
       * @param {String/Number/Mask} mask The source mask
       *
       * @return {Address} Returns a new {@link Mask} object if `mask`
       * is a `String` or a `Number`, or the original {@link Mask} object otherwise
       */
      getMaskAsObject: function(mask) {
        return mask instanceof Mask ? mask : new Mask(mask);
      }

    }, {

      /**
       * Create a new {@link Subnet IP v4 Subnet}
       *
       * @param {Number/String/Address} network The subnet network address
       * @param {Number/String/Mask}    network The subnet mask
       * @constructor
       */
      init: function(network, mask) {
        this.setNetwork(Subnet.getAddressAsObject(network));
        this.setMask   (Subnet.getMaskAsObject(mask));

        // Check the network/mask validity
        var networkAddress;
        if((networkAddress = this.getMask().apply(this.getNetwork()).getAddress()) != this.getNetwork().getAddress()) {
          throw new Error('Invalid IPv4 Subnet "' + this + '", should be "' + new Subnet(networkAddress, mask) + '".');
        }
      },

      /**
       * Get the String representation of the Subnet in the following format:
       *
       *     "<network>/<mask size>"
       *
       * where `network` is the {@link Address#toString String representation of the subnet network address}
       * and `mask size` is the {@link Address#getMaskSize mask size of the subnet mask}
       *
       * Example:
       *
       *     var s = new Subnet('192.168.1.0', '255.255.255.0');
       *     s.toString() // "192.168.1.0/24"
       *
       * @return {String} Return the String representation of the Subnet
       */
      toString: function() {
        return this.getNetwork().toString() + '/' + this.getMask().getMaskSize();
      },

      /**
       * Get the broadcast address of this subnet
       *
       * @return {Address} Returns the broadcast address of this subnet
       */
      getBroadcastAddress: function() {
        return new Address(this.getNetwork().getAddress() | ~this.getMask().getMask());
      },

      /**
       * Determine whether an address is the network address of this subnet or not
       *
       * @param {Number/String/Address} address The address to check
       * @return {Boolean} Returns `true` if `address` is the network address of this subnet, `false` otherwise
       */
      isNetworkAddress: function(address) {
        return Subnet.getAddressAsObject(address).getAddress() === this.getNetwork().getAddress();
      },

      /**
       * Determine whether an address is the broadcast address of this subnet or not
       *
       * @param {Number/String/Address} address The address to check
       * @return {Boolean} Returns `true` if `address` is the broadcast address of this subnet, `false` otherwise
       */
      isBroadcastAddress: function(address) {
        return Subnet.getAddressAsObject(address).getAddress() === this.getBroadcastAddress().getAddress();
      },

      /**
       * Determine whether an IP address is valid within this subnet, or not. A valid
       * IP address matches the following conditions:
       *
       *    1. the address belongs to this subnet
       *    2. the address IS NOT the network address (reserved)
       *    3. the address IS NOT the broadcast address (reserved)
       *
       * @param {Number/String/Address} address The address to check
       * @return {Boolean} Returns `true` if the address is valid, `false` otherwise.
       */
      isValidAddress: function(address) {
        var tmpAddress = Subnet.getAddressAsObject(address);
        return this.getMask().apply(tmpAddress).getAddress() === this.getNetwork().getAddress() &&
          !this.isNetworkAddress(tmpAddress) &&
          !this.isBroadcastAddress(tmpAddress);
      }

    });

  return Subnet;
});
